//
//  Installation.swift
//  TestBsonLeak
//
//  Created by Nikolay Dzhulay on 03/01/2021.
//

import Foundation

public struct Installation: Codable {

    public enum DeviceType: String, Codable {
        case ios
        case android
    }

    // MARK: CosmosDbDocument

    /// To generate this 2 values use CompoundId, `partitionKey` is year, `id` is `CompoundId.string`
    public let id: String // Format will be "UUID_yyyy-mm", where "yyyy" is a partition key
    public let partitionKey: String // "yyyy", year when installation registered

    public let appInfo: AppInfo
    public let deviceInfo: DeviceInfo
    public let installationInfo: InstallationInfo

    public let createdAt: Date
    public let createdAtSeconds: Int64

    public init(id: String,
                partitionKey: String,
                appInfo: AppInfo,
                deviceInfo: DeviceInfo,
                installationInfo: InstallationInfo,
                createdAt: Date) {
        self.id = id
        self.partitionKey = partitionKey
        self.appInfo = appInfo
        self.deviceInfo = deviceInfo
        self.installationInfo = installationInfo
        self.createdAt = createdAt
        self.createdAtSeconds = Int64(createdAt.timeIntervalSince1970.rounded(.up))
    }

    public enum CodingKeys: CodingKey {
        case id
        case partitionKey
        case createdAt
        case createdAtSeconds

        case appInfo
        case deviceInfo
        case installationInfo
    }
}

extension Installation {
    public struct AppInfo: Codable, Equatable {

        public let isDebug: Bool

        /// app info
        public let appIdentifier: String
        public let appVersion: String

        public init(isDebug: Bool,
                    appIdentifier: String,
                    appVersion: String) {
            self.isDebug = isDebug
            self.appIdentifier = appIdentifier
            self.appVersion = appVersion
        }

        // MARK: Private

        fileprivate enum CodingKeys: String, CodingKey {
            case appIdentifier
            case appVersion
            case isDebug
        }
    }
}

extension Installation {
    public struct DeviceInfo: Codable, Equatable {
        public let osVersion: String
        public let deviceType: DeviceType
        public let device: String

        public let localeIdentifier: String
        public let timeZone: String

        public init(osVersion: String,
                    deviceType: DeviceType,
                    device: String,
                    localeIdentifier: String,
                    timeZone: String) {
            self.osVersion = osVersion
            self.deviceType = deviceType
            self.device = device
            self.localeIdentifier = localeIdentifier
            self.timeZone = timeZone
        }

        // MARK: Private

        fileprivate enum CodingKeys: String, CodingKey {
            case osVersion
            case deviceType
            case device
            case localeIdentifier
            case timeZone
        }
    }
}

extension Installation {
    public struct InstallationInfo: Codable, Equatable {

        // push tokens
        public let deviceToken: String?
        public let fcmToken: String?

        public let tags: [String]
        public let cloudKitUserId: String?

        public let languageCode: String

        public init(deviceToken: String?,
                    fcmToken: String?,
                    tags: [String],
                    cloudKitUserId: String?,
                    languageCode: String) {
            self.deviceToken = deviceToken
            self.fcmToken = fcmToken
            self.tags = tags
            self.cloudKitUserId = cloudKitUserId
            self.languageCode = languageCode
        }

        // MARK: Private

        fileprivate enum CodingKeys: String, CodingKey {
            case deviceToken
            case fcmToken
            case tags
            case cloudKitUserId
            case languageCode
        }
    }
}

extension Installation  {

    /// This compound id should help find object faster, because we know partition
    /// For document id we should use `CompoundId.string` for partitionKey `CompoundId.partitionKey`
    public struct CompoundId: Codable {
        public let uniqId: String
        public let year: Int

        public var string: String {
            return uniqId + CompoundId.separator + partitionKey
        }

        public var partitionKey: String {
            return String(format: "%04d", year)
        }

        public init(id: String, date: Date) {

            self.uniqId = id

            let calendar = Calendar.init(identifier: Calendar.Identifier.gregorian)
            self.year = calendar.component(.year, from: date)
        }

        public init(string: String) throws {
            let components = string.components(separatedBy: CompoundId.separator)
            guard components.count == 2, !components[0].isEmpty,
                  let year = Int(components[1]), (2018...2099).contains(year) else {
                throw Errors.cantCreateCompoundId(string)
            }
            self.uniqId = components[0]
            self.year = year
        }

        // MARK: Codable
        public init(from decoder: Decoder) throws {
            let stringContainer = try decoder.singleValueContainer()
            let string = try stringContainer.decode(String.self)
            try self.init(string: string)
        }

        public func encode(to encoder: Encoder) throws {
            var container = encoder.singleValueContainer()
            try container.encode(string)
        }

        // MARK: Private
        private static let separator = "_"

    }
}

private extension Installation.CompoundId {
    enum Errors: LocalizedError {

        case cantCreateCompoundId(String)

        // MARK: EnumError

        var errorDescription: String? {
            switch self {
            case let .cantCreateCompoundId(string):
                return "Can't create compound id from \(string) (Installation.CompoundId.Errors.cantCreateCompoundId)"
            }
        }
    }
}

